# -*- coding: utf-8 -*-
import time
import json


class SquareService(object):


    SQUARE_EXCEPTION = {
        'code': 1,
        'message': 3,
        'metadata': 2
    }

    def __init__(self):
        pass

    def inviteIntoSquareChat(self, inviteeMids, squareChatMid):
        sqrd = [128, 1, 0, 1, 0, 0, 0, 20, 105, 110, 118, 105, 116, 101, 73, 110,
                116, 111, 83, 113, 117, 97, 114, 101, 67, 104, 97, 116, 0, 0, 0, 0]
        sqrd += [12, 0, 1]
        sqrd += [15, 0, 1, 11, 0, 0, 0, len(inviteeMids)]
        for mid in inviteeMids:
            sqrd += [0, 0, 0, len(mid)]
            for value in mid:
                sqrd.append(ord(value))
        sqrd += [11, 0, 2] + self.getStringBytes(squareChatMid)
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, baseException=SquareService.SQUARE_EXCEPTION)

    def inviteToSquare(self, squareMid, invitees, squareChatMid):
        sqrd = [128, 1, 0, 1] + \
            self.getStringBytes("inviteToSquare") + [0, 0, 0, 0]
        sqrd += [12, 0, 1]
        sqrd += [11, 0, 2] + self.getStringBytes(squareMid)
        sqrd += [15, 0, 3, 11, 0, 0, 0, len(invitees)]
        for mid in invitees:
            sqrd += [0, 0, 0, len(mid)]
            for value in mid:
                sqrd.append(ord(value))
        sqrd += [11, 0, 4] + self.getStringBytes(squareChatMid)
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, baseException=SquareService.SQUARE_EXCEPTION)

    def getJoinedSquares(self, continuationToken=None, limit=50):
        sqrd = [128, 1, 0, 1, 0, 0, 0, 16, 103, 101, 116, 74, 111, 105,
                110, 101, 100, 83, 113, 117, 97, 114, 101, 115, 0, 0, 0, 0]
        sqrd += [12, 0, 1]
        #sqrd += [11, 0, 2] + self.getStringBytes(continuationToken)
        sqrd += [8, 0, 3] + self.getIntBytes(limit)
        sqrd += [0, 0]
        #sqr_rd = sqrd
        #data = bytes(sqr_rd)
        #res = self.req_h2.post(self.LINE_HOST_DOMAIN + self.LINE_SQUARE_QUERY_PATH, data=data, headers=self.square_headers)
        #data = res.content
        # return self.tryReadData(data, mode=0)['getJoinedSquares']
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, baseException=SquareService.SQUARE_EXCEPTION)

    def markAsRead(self, squareChatMid, messageId):
        sqrd = [128, 1, 0, 1, 0, 0, 0, 10, 109, 97, 114,
                107, 65, 115, 82, 101, 97, 100, 0, 0, 0, 0]
        sqrd += [12, 0, 1]
        sqrd += [11, 0, 2, 0, 0, 0, len(squareChatMid)]
        for value in squareChatMid:
            sqrd.append(ord(value))
        sqrd += [11, 0, 4, 0, 0, 0, len(messageId)]
        for value in messageId:
            sqrd.append(ord(value))
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, baseException=SquareService.SQUARE_EXCEPTION)

    def reactToMessage(self, squareChatMid, messageId, reactionType=2):
        """
        - reactionType
            ALL     = 0,
            UNDO    = 1,
            NICE    = 2,
            LOVE    = 3,
            FUN     = 4,
            AMAZING = 5,
            SAD     = 6,
            OMG     = 7,
        """
        sqrd = [128, 1, 0, 1] + \
            self.getStringBytes("reactToMessage") + [0, 0, 0, 0]
        sqrd += [12, 0, 1]
        sqrd += [8, 0, 1] + self.getIntBytes(0)  # reqSeq
        sqrd += [11, 0, 2] + self.getStringBytes(squareChatMid)
        sqrd += [11, 0, 3] + self.getStringBytes(messageId)
        sqrd += [8, 0, 4] + self.getIntBytes(reactionType)
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, baseException=SquareService.SQUARE_EXCEPTION)

    def findSquareByInvitationTicket(self, invitationTicket):
        sqrd = [128, 1, 0, 1] + \
            self.getStringBytes("findSquareByInvitationTicket") + [0, 0, 0, 0]
        sqrd += [12, 0, 1]
        sqrd += [11, 0, 2] + self.getStringBytes(invitationTicket)
        sqrd += [0, 0]
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, baseException=SquareService.SQUARE_EXCEPTION)

    def fetchMyEvents(self, subscriptionId: int=0, syncToken: str=None, continuationToken: str=None, limit: int=100):
        params = [
            [12, 1, [
                [10, 1, subscriptionId],
                [11, 2, syncToken],
                [8, 3, limit],
                [11, 4, continuationToken]
            ]]
        ]
        sqrd = self.generateDummyProtocol('fetchMyEvents', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, baseException=SquareService.SQUARE_EXCEPTION)

    def fetchSquareChatEvents(self, squareChatMid, syncToken='', subscriptionId=0, limit=100):
        params = [
            [12, 1, [
                [10, 1, subscriptionId],
                [11, 2, squareChatMid],
                # [11, 3, syncToken],
                [8, 4, limit],
                [8, 5, 1], # direction
                [8, 6, 1], # inclusive
                # [11, 7, continuationToken]
            ]]
        ]
        sqrd = self.generateDummyProtocol('fetchSquareChatEvents', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, baseException=SquareService.SQUARE_EXCEPTION)

    def sendSquareMessage(self, squareChatMid: str, text: str, contentType: int=0, contentMetadata: dict={}, relatedMessageId: str=None):
        message = [
            # [11, 1, _from],
            [11, 2, squareChatMid],
            [10, 5, 0],
            [10, 6, 0],
            [11, 10, text],
            [2, 14, False], # hasContent
            [8, 15, contentType],  # contentType
            [13, 18, [11, 11, contentMetadata]],
            [8, 19, 0],
        ]
        if relatedMessageId is not None:
            message.append(
                [11, 21, relatedMessageId]
            )
            message.append(
                [8, 22, 3] # messageRelationType; FORWARD(0), AUTO_REPLY(1), SUBORDINATE(2), REPLY(3);

            )
            message.append(
                [8, 24, 2] # relatedMessageServiceCode; 1 for Talk 2 for Square
            )
        params = [
            [12, 1, [
                [8, 1, 0],  # reqSeq
                [11, 2, squareChatMid],
                [12, 3, [
                    [12, 1, message],
                    [8, 3, 4],  # fromType
                    # [10, 4, 0],  # squareMessageRevision
                    # [8, 5, 1]
                ]],
            ]]
        ]
        sqrd = self.generateDummyProtocol('sendMessage', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, encType=0, baseException=SquareService.SQUARE_EXCEPTION)

    def sendSquareTextMessage(self, squareChatMid: str, text: str, contentMetadata: dict={}, relatedMessageId: str=None):
        return self.sendSquareMessage(squareChatMid, text, 0, contentMetadata, relatedMessageId)

    def getSquare(self, squareMid):
        params = [
            [12, 1, [
                [11, 2, squareMid]
            ]]
        ]
        sqrd = self.generateDummyProtocol('getSquare', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, encType=0, baseException=SquareService.SQUARE_EXCEPTION)

    def getJoinableSquareChats(self, squareMid, continuationToken=None):
        params = [
            [12, 1, [
                [11, 2, squareMid],
                [11, 10, continuationToken],
                [8, 11, 100]
            ]]
        ]
        sqrd = self.generateDummyProtocol('getJoinableSquareChats', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, encType=0, baseException=SquareService.SQUARE_EXCEPTION)

    def createSquare(self, name: str, displayName: str, profileImageObsHash: str = "0h6tJf0hQsaVt3H0eLAsAWDFheczgHd3wTCTx2eApNKSoefHNVGRdwfgxbdgUMLi8MSngnPFMeNmpbLi8MSngnPFMeNmpbLi8MSngnOA", desc: str = "test with CHRLINE API", searchable: bool = True, SquareJoinMethodType: int = 0):
        """
        - SquareJoinMethodType
            NONE(0),
            APPROVAL(1),
            CODE(2);
        """
        params = [
            [12, 1, [
                [8, 2, self.getCurrReqId()],
                [12, 2, [
                    [11, 2, name],
                    [11, 4, profileImageObsHash],
                    [11, 5, desc],
                    [2, 6, searchable],
                    [8, 7, 1],  # type
                    [8, 8, 1],  # categoryId
                    [10, 10, 0],  # revision
                    [2, 11, True],  # ableToUseInvitationTicket
                    [12, 14, [
                        [8, 1, SquareJoinMethodType]
                    ]],
                    [2, 15, False],  # adultOnly
                    [15, 16, [11, []]]  # svcTags
                ]],
                [12, 3, [
                    [11, 3, displayName],
                    # [11, 4, profileImageObsHash],
                    [2, 5, True],  # ableToReceiveMessage
                    [10, 9, 0],  # revision
                ]]
            ]]
        ]
        sqrd = self.generateDummyProtocol('createSquare', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, encType=0, baseException=SquareService.SQUARE_EXCEPTION)

    def getSquareChatAnnouncements(self, squareMid: str):
        params = [
            [12, 1, [
                [11, 2, squareMid],
            ]]
        ]
        sqrd = self.generateDummyProtocol(
            'getSquareChatAnnouncements', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, encType=0, baseException=SquareService.SQUARE_EXCEPTION)

    def leaveSquareChat(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("leaveSquareChat is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "leaveSquareChat", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareChatMember(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareChatMember is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareChatMember", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def searchSquares(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("searchSquares is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "searchSquares", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def updateSquareFeatureSet(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("updateSquareFeatureSet is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "updateSquareFeatureSet", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def joinSquare(self, squareMid, displayName, ableToReceiveMessage: bool=False, passCode: str=None):
        params = [
            [12, 1, [
                [11, 2, squareMid],
                [12, 3, [
                    [11, 2, squareMid],
                    [11, 3, displayName],
                    [2, 5, ableToReceiveMessage],
                    [10, 9, 0],
                ]],
                [12, 5, [
                    [12, 2, [
                        [11, 1, passCode]
                    ]]
                ]]
            ]]
        ]
        sqrd = self.generateDummyProtocol('joinSquare', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, encType=0, baseException=SquareService.SQUARE_EXCEPTION)

    def getSquarePopularKeywords(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getPopularKeywords is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getPopularKeywords", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def reportSquareMessage(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("reportSquareMessage is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "reportSquareMessage", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def updateSquareMemberRelation(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("updateSquareMemberRelation is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "updateSquareMemberRelation", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def leaveSquare(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("leaveSquare is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "leaveSquare", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareMemberRelations(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareMemberRelations is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareMemberRelations", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def removeSquareSubscriptions(self, subscriptionIds: list=[]):
        params = [
            [12, 1, [
                [15, 2, [10, subscriptionIds]],
            ]]
        ]
        sqrd = self.generateDummyProtocol('removeSquareSubscriptions', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, encType=0, baseException=SquareService.SQUARE_EXCEPTION)

    def getSquareMembers(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareMembers is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareMembers", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def updateSquareChat(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("updateSquareChat is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "updateSquareChat", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareMessageReactions(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getMessageReactions is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getMessageReactions", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def destroySquareMessage(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("destroyMessage is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "destroyMessage", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def reportSquareChat(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("reportSquareChat is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "reportSquareChat", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def unsendSquareMessage(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("unsendMessage is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "unsendMessage", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def deleteSquareChatAnnouncement(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("deleteSquareChatAnnouncement is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "deleteSquareChatAnnouncement", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def createSquareChat(self, squareChatMid: str, name: str, chatImageObsHash: str, squareChatType: int = 1, maxMemberCount: int = 5000, ableToSearchMessage: int = 1, squareMemberMids: list = []):
        """
        - SquareChatType:
            OPEN(1),
            SECRET(2),
            ONE_ON_ONE(3),
            SQUARE_DEFAULT(4);
        - ableToSearchMessage:
            NONE(0),
            OFF(1),
            ON(2);
        """
        params = [
            [12, 1,  [
                [8, 1, self.getCurrReqId()],
                [12, 2, [
                    [11, 1, squareChatMid],
                    [8, 3, squareChatType],
                    [11, 4, name],
                    [11, 5, chatImageObsHash],
                    [8, 7, maxMemberCount],
                    [8, 11, ableToSearchMessage]
                ]],
                [15, 3, [11, squareMemberMids]]
            ]]
        ]
        sqrd = self.generateDummyProtocol(
            "createSquareChat", params, self.SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.SquareService_API_PATH, sqrd, self.SquareService_RES_TYPE)

    def deleteSquareChat(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("deleteSquareChat is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "deleteSquareChat", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareChatMembers(self, squareChatMid: str, continuationToken: str = None, limit: int = 200):
        GetSquareChatMembersRequest = [
            [11, 1, squareChatMid],
            [8, 3, limit]
        ]
        if continuationToken is not None:
            GetSquareChatMembersRequest.append(
                [11, 2, continuationToken])
        params = [
            [12, 1, GetSquareChatMembersRequest]
        ]
        sqrd = self.generateDummyProtocol(
            "getSquareChatMembers", params, self.SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.SquareService_API_PATH, sqrd, self.SquareService_RES_TYPE)

    def getSquareFeatureSet(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareFeatureSet is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareFeatureSet", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def updateSquareAuthority(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("updateSquareAuthority is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "updateSquareAuthority", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def rejectSquareMembers(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("rejectSquareMembers is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "rejectSquareMembers", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def deleteSquare(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("deleteSquare is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "deleteSquare", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def reportSquare(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("reportSquare is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "reportSquare", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareInvitationTicketUrl(self, mid: str):
        params = [
            [12, 1, [
                [11, 2, mid],
            ]]
        ]
        sqrd = self.generateDummyProtocol('getInvitationTicketUrl', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, encType=0, baseException=SquareService.SQUARE_EXCEPTION)

    def updateSquareChatMember(self, squareMemberMid: str, squareChatMid: str, notificationForMessage: bool=True, notificationForNewMember: bool=True, updatedAttrs: list=[6]):
        """
        - SquareChatMemberAttribute:
            MEMBERSHIP_STATE(4),
            NOTIFICATION_MESSAGE(6),
            NOTIFICATION_NEW_MEMBER(7);
        """
        params = [
            [12, 1, [
                [14, 2, [8, updatedAttrs]],
                [12, 3, [
                    [11, 1, squareMemberMid],
                    [11, 2, squareChatMid],
                    [2, 5, notificationForMessage],
                    [2, 6, notificationForNewMember],
                ]],
            ]]
        ]
        sqrd = self.generateDummyProtocol('updateSquareChatMember', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, encType=0, baseException=SquareService.SQUARE_EXCEPTION)

    def updateSquareMember(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("updateSquareMember is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "updateSquareMember", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def updateSquare(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("updateSquare is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "updateSquare", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareAuthorities(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareAuthorities is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareAuthorities", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def updateSquareMembers(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("updateSquareMembers is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "updateSquareMembers", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareChatStatus(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareChatStatus is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareChatStatus", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def approveSquareMembers(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("approveSquareMembers is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "approveSquareMembers", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareStatus(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareStatus is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareStatus", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def searchSquareMembers(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("searchSquareMembers is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "searchSquareMembers", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def checkSquareJoinCode(self, squareMid: str, code: str):
        params = [
            [12, 1, [
                [11, 2, squareMid],
                [11, 3, code],
            ]]
        ]
        sqrd = self.generateDummyProtocol('checkJoinCode', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, encType=0, baseException=SquareService.SQUARE_EXCEPTION)

    def createSquareChatAnnouncement(self, squareChatMid: str, messageId: str, text: str, senderSquareMemberMid: str, createdAt: int, announcementType: int = 0):
        """
        - SquareChatAnnouncementType:
            TEXT_MESSAGE(0);
        """
        params = [
            [12, 1, [
                [8, 1, self.getCurrReqId()],
                [11, 2, squareChatMid],
                [12, 3, [
                    [8, 2, announcementType],
                    [12, 3, [
                        [12, 1, [
                            [11, 1, messageId],
                            [11, 2, text],
                            [11, 3, senderSquareMemberMid],
                            [10, 4, createdAt]
                        ]]
                    ]]
                ]]
            ]]
        ]
        sqrd = self.generateDummyProtocol(
            "createSquareChatAnnouncement", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareAuthority(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareAuthority is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareAuthority", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareChat(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareChat is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareChat", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def refreshSquareSubscriptions(self, subscriptions: list):
        params = [
            [12, 1, [
                [15, 2, [10, subscriptions]],
            ]]
        ]
        sqrd = self.generateDummyProtocol('refreshSquareSubscriptions', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, encType=0, baseException=SquareService.SQUARE_EXCEPTION)

    def getJoinedSquareChats(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getJoinedSquareChats is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getJoinedSquareChats", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def joinSquareChat(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("joinSquareChat is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "joinSquareChat", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def findSquareByEmid(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("findSquareByEmid is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "findSquareByEmid", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareMemberRelation(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareMemberRelation is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getSquareMemberRelation", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareMember(self, squareMemberMid: str):
        params = [
            [12, 1, [
                [11, 1, squareMemberMid],
            ]]
        ]
        sqrd = self.generateDummyProtocol('getSquareMember', params, 4)
        return self.postPackDataAndGetUnpackRespData(self.LINE_SQUARE_ENDPOINT, sqrd, 4, encType=0, baseException=SquareService.SQUARE_EXCEPTION)

    def destroySquareMessages(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("destroyMessages is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "destroyMessages", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareCategories(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getCategories is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getCategories", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def reportSquareMember(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("reportSquareMember is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "reportSquareMember", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def getSquareNoteStatus(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getNoteStatus is not implemented")
        params = []
        sqrd = self.generateDummyProtocol(
            "getNoteStatus", params, SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(SquareService_API_PATH, sqrd, SquareService_RES_TYPE)

    def searchSquareChatMembers(self, squareChatMid: str, displayName: str = '', continuationToken: str = None, limit: int = 20):
        SearchSquareChatMembersRequest = [
            [11, 1, squareChatMid],
            [12, 2, [
                [11, 1, displayName],
                # [2, 2, True]  # includingMe
            ]],
            [8, 4, limit],
        ]
        if continuationToken is not None:
            SearchSquareChatMembersRequest.append(
                [11, 3, continuationToken])
        params = [
            [12, 1, SearchSquareChatMembersRequest]
        ]
        sqrd = self.generateDummyProtocol(
            "searchSquareChatMembers", params, self.SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.SquareService_API_PATH, sqrd, self.SquareService_RES_TYPE)


    def getSquareChatFeatureSet(self):
        """
        AUTO_GENERATED_CODE! DONT_USE_THIS_FUNC!!
        """
        raise Exception("getSquareChatFeatureSet is not implemented")
        params = []
        sqrd = self.generateDummyProtocol("getSquareChatFeatureSet", params, self.SquareService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.SquareService_API_PATH ,sqrd,  self.SquareService_RES_TYPE)


# -*- coding: utf-8 -*-
import json
import time
import hashlib
from base64 import b64encode
import uuid
import urllib
import requests
import httpx


class Object(object):

    def __init__(self):
        self.Hraders4Obs = {
            'User-Agent': self.server.Headers['User-Agent'],
            'X-Line-Access': self.acquireEncryptedAccessToken().split('\x1e')[1],
            'X-Line-Application': self.server.Headers['x-line-application'],
            'X-Line-Mid': self.mid,
            'x-lal': self.LINE_LANGUAGE,
        }
        self.obsConn = httpx.Client(http2=True, timeout=None)
    """ 
    
    TimelineObs 
    Source: https://github.com/DeachSword/LINE-DemoS-Bot/blob/master/api/obs/obs.py
    
    """

    def updateProfileImage(self, path, storyShare=False, type='p', mid: str = None):
        """
        use `mid` for square member
        """
        if mid is None:
            mid = self.mid
        midType = self.getToType(mid)
        if midType == 0:
            obsPath = f'talk/p/{mid}'
        elif midType == 5:
            obsPath = f'g2/member/{mid}'
        else:
            raise ValueError(f"Invalid midType: {midType}")
        oType = 'IMAGE'
        params = {}
        is_video = self.checkIsVideo(path)
        hstr = 'DearSakura_%s' % int(time.time() * 1000)
        hstr = hashlib.md5(hstr.encode()).hexdigest()
        filename = "profile.jpg"
        metaData = {
            "profileContext": {
                "storyShare": storyShare
            }
        }
        if type == 'vp':
            params["cat"] = "vp.mp4"  # let server know u have vp.mp4
        is_video = self.checkIsVideo(path)
        if is_video:
            obsPath = f'talk/vp/{mid}'
            oType = "VIDEO"
            filename = f"{hstr}.mp4"
            params = {
                "cat": "vp.mp4",
            }
        talkMeta = b64encode(json.dumps(
            metaData).encode('utf-8')).decode('utf-8')
        try:
            objId, objHash, respHeaders = self.uploadObjectForService(
                path, oType, obsPath, "1341209850", params, talkMeta, filename)
        except Exception as e:
            raise Exception(
                f"updateProfileImage failure: {e}")
        return True

    def updateProfileCover(self, path, storyShare=False):
        hstr = 'DearSakura_%s' % int(time.time() * 1000)
        objId = hashlib.md5(hstr.encode()).hexdigest()
        obsPath = f'myhome/c/{objId}'
        oType = 'IMAGE'
        filename = f"{objId}.jpg"
        params = None
        is_video = self.checkIsVideo(path)
        if is_video:
            obsPath = f'myhome/vc/{objId}'
            oType = 'VIDEO'
            filename = f"{objId}.mp4"
            params = {
                "cat": 'vp.mp4'
            }
        try:
            objId, objHash, respHeaders = self.uploadObjectForService(
                path, oType, obsPath, "1341209850", params, filename=filename)
        except Exception as e:
            raise Exception(e)
        if is_video:
            _url, _vc_url, _objId, _vc_objId = self.getProfileCoverObjIdAndUrl(
                self.mid)
            home = self.updateProfileCoverById(
                _objId, objId, storyShare=storyShare)
        else:
            home = self.updateProfileCoverById(objId, storyShare=storyShare)
        return home

    def updateChatProfileImage(self, groupId: str, path: str, oType: str = 'IMAGE'):
        midType = self.getToType(groupId)
        if midType in [1, 2]:
            obsPath = f'talk/g/{groupId}'
        elif midType == 3:
            obsPath = f'g2/group/{groupId}'
        elif midType == 4:
            obsPath = f'g2/chat/{groupId}'
        else:
            raise ValueError(f"Invalid midType: {midType}")
        params = {
            'cat': 'original'
        }
        try:
            objId, objHash, respHeaders = self.uploadObjectForService(
                path, oType, obsPath, "1341209850", params, filename='profile.jpg')
        except Exception as e:
            raise Exception(
                f"updateChatProfileImage failure: {e}")
        return True

    def updateImage2Album(self, homeId: str, albumId: int, pathOrBytes: any, oType: str = "IMAGE", updateAlbum: bool = True):
        """
        Upload Object to Album Obs

        will use `updateAlbum` to update the album, if content uploads success. 
        ---------------
        Return `OBS_OID` and `OBS_HASH`
        """
        oids = []
        additionalHeaders = {
            'X-Line-Album': str(albumId),
            'X-Line-Mid': homeId,
        }
        if type(pathOrBytes) != list:
            pathOrBytes = [pathOrBytes]
        for _uploadData in pathOrBytes:
            oid = f"{uuid.uuid1().hex}.20{ time.strftime('%m%d') }08"
            params = {
                "oid": oid
            }
            obsObjId, obsHash, respHeaders = self.uploadObjectForService(
                _uploadData, oType, f'album/a/{oid}', issueToken4ChannelId="1341209850", params=params, additionalHeaders=additionalHeaders)
            oids.append(obsObjId)
        if updateAlbum:
            self.addImageToAlbum(homeId, albumId, oids)
        return oids

    def uploadObjHome(self, path, type='image', objId=None):
        if type not in ['image', 'video', 'audio']:
            raise Exception('Invalid type value')
        if type == 'image':
            contentType = 'image/jpeg'
        elif type == 'video':
            contentType = 'video/mp4'
        elif type == 'audio':
            contentType = 'audio/mp3'
        if not objId:
            hstr = 'DearSakura_%s' % int(time.time()*1000)
            objId = hashlib.md5(hstr.encode()).hexdigest()
        file = open(path, 'rb').read()
        params = {
            'name': f"{objId}.mp4",
            'oid': '%s' % str(objId),
            'type': type,
            'ver': '2.0'  # 2.0 :p
        }
        hr = self.server.additionalHeaders(self.server.timelineHeaders, {
            'Content-Type': contentType,
            'Content-Length': str(len(file)),
            'x-obs-params': self.genOBSParams(params, 'b64')  # base64 encode
        })
        r = self.server.postContent(
            self.LINE_OBS_DOMAIN + '/myhome/h/upload.nhn', headers=hr, data=file)
        if r.status_code != 201:
            raise Exception(
                f"Upload object home failure. Receive statue code: {r.status_code}")
        return objId

    def uploadMultipleImageToTalk(self, paths: list, to: str, oTypes: list = None):
        if type(paths) != list:
            raise Exception('paths must be list')
        if oTypes is None:
            oTypes = ["IMAGE" for _ in range(len(paths))]
        oType = iter(oTypes)
        sqrd_base = [13, 0, 18, 11, 11]
        hashmap = {
            "GID": "0",
            "GSEQ": "1",
            "GTOTAL": str(len(paths))
        }
        sqrd = sqrd_base + self.getIntBytes(len(hashmap.keys()))
        for hm in hashmap.keys():
            sqrd += self.getStringBytes(hm)
            sqrd += self.getStringBytes(hashmap[hm])
        sqrd += [0]
        data = bytes(sqrd)
        msg = json.dumps({
            "message": b64encode(data).decode('utf-8')
        })
        talkMeta = b64encode(msg.encode('utf-8')).decode('utf-8')
        res = self.uploadObjTalk(
            paths[0], next(oType), to=to, talkMeta=talkMeta, returnHeaders=True)
        gid = res['x-line-message-gid']
        msgIds = [res['x-obs-oid']]
        if len(paths) > 1:
            hashmap["GID"] = gid
            nc = 2
            for img in paths[1:]:
                self.log(f"Upload image-{nc} with GID-{gid}...", True)
                hashmap["GSEQ"] = str(nc)
                sqrd = sqrd_base + self.getIntBytes(len(hashmap.keys()))
                for hm in hashmap.keys():
                    sqrd += self.getStringBytes(hm)
                    sqrd += self.getStringBytes(hashmap[hm])
                sqrd += [0]
                data = bytes(sqrd)
                msg = json.dumps({
                    "message": b64encode(data).decode('utf-8')
                })
                talkMeta = b64encode(msg.encode('utf-8')).decode('utf-8')
                res = self.uploadObjTalk(
                    img, next(oType), to=to, talkMeta=talkMeta, returnHeaders=True)
                msgIds.append(res['x-obs-oid'])
                nc += 1
        return gid

    def uploadObjTalk(self, pathOrBytes, oType='image', objId=None, to=None, talkMeta=None, returnHeaders=False, filename: str = None, isOriginal: bool = False):
        if oType.lower() not in ['image', 'gif', 'video', 'audio', 'file']:
            raise Exception('Invalid type value')
        # url = self.LINE_OBS_DOMAIN + '/talk/m/upload.nhn' #if reqseq not working
        # url = self.LINE_OBS_DOMAIN + '/r/talk/m/reqseq'
        serviceName = 'talk'
        obsNamespace = 'm'
        params = {}
        if oType.lower() == 'gif':
            params['cat'] = 'original'
            oType = 'IMAGE'
        if isOriginal:
            params['cat'] = 'original'
        if to != None:
            params.update({
                "oid": "reqseq",
                "reqseq": str(self.getCurrReqId()),
                "tomid": to
            })
            objId = "reqseq"
            if self.getToType(to) == 4:
                serviceName = 'g2'
        else:
            if objId != None:
                params['oid'] = objId
        obsObjId, obsHash, respHeaders = self.uploadObjectForService(
            pathOrBytes, oType, f'{serviceName}/{obsNamespace}/{objId}', params=params, talkMeta=talkMeta, filename=filename)
        if returnHeaders:
            return respHeaders
        objId = obsObjId
        objHash = obsHash  # for view on cdn
        return objId

    def uploadStoryObject(self, pathOrBytes: any, oType='IMAGE'):
        """
        Upload Object to Story Obs

        ---------------
        Return `OBS_OID` and `OBS_HASH`
        """
        oid = f"{uuid.uuid1().hex}tffffffff"
        params = {
            "oid": oid
        }
        obsObjId, obsHash, respHeaders = self.uploadObjectForService(
            pathOrBytes, oType, f'story/st/{oid}', issueToken4ChannelId="1341209850", params=params)
        return obsObjId, obsHash

    def forwardObjectMsg(self, to: str, msgId: str, _from: str = 'u', **kwargs):
        """
        Forward object to Talk

        use `_from` to specify the source mid of the messageid
        """
        copyFromService = 'talk'
        copyFromSid = 'm'
        copy2Service = 'talk'
        copy2Sid = 'm'
        data = {
            'oid': 'reqseq',
            'tomid': to,
            'reqseq': self.getCurrReqId(),
            'ver': '1.0',
        }
        toType = self.getToType(to)
        if toType in [0, 1, 2]:
            pass
        elif toType == 4:
            copy2Service = 'g2'
        else:
            raise ValueError(f"Invalid `to` midType: {toType}")
        toType = self.getToType(_from)
        if toType in [0, 1, 2]:
            pass
        elif toType == 4:
            copyFromService = 'g2'
        else:
            raise ValueError(f"Invalid `_from` midType: {toType}")
        kwargs['copyFromService'] = copyFromService
        kwargs['copyFromSid'] = copyFromSid
        kwargs['copyFromObjId'] = msgId
        kwargs['copy2Service'] = copy2Service
        kwargs['copy2Sid'] = copy2Sid
        kwargs['copy2ObjId'] = 'reqseq'
        return self.forwardObjectForService(data, **kwargs)

    def forwardKeepObjectMsg(self, to, oid, contentId, contentType='image'):
        # SHARE KEEP CONTENTS V1
        data = {
            'copyFrom': f'/keep/p/{oid}',
            'tomid': to,
            'reqseq': self.getCurrReqId(),
            'type': contentType,
            'contentId': contentId,
            'size': '-1',
        }
        if contentType == 'image':
            data['cat'] = 'original'
        elif contentType in ['audio', 'video']:
            data['duration'] = '3000'
        print(self.server.timelineHeaders)
        r = self.server.postContent(
            self.LINE_HOST_DOMAIN + '/oa/r/talk/m/reqseq/copy.nhn', data=data, headers=self.server.timelineHeaders)
        if not self.checkRespIsSuccessWithLpv(r):
            raise Exception(f'Forward object failure: {r.status_code}')
        return r.headers['X-Obs-Oid']

    def forwardKeepContent2Mid(self, to: str, oid: str, contentId: str, **kwargs):
        # SHARE KEEP CONTENTS V2
        data = {
            'tomid': to,
            'reqseq': self.getCurrReqId(),
        }
        kwargs['copyFromService'] = 'keep'
        kwargs['copyFromSid'] = 'p'
        kwargs['copyFromObjId'] = oid
        kwargs['contentId'] = contentId
        return self.forwardObjectForService(data, **kwargs)

    def trainingImage(self, msgId):
        data = {
            "useAsTrainingData": False,
            "input": 'obs://talk/m/%s' % msgId,
            "filters": {
                "OCR": {
                    "params": {
                        "useColorPicker": "true"
                    }
                }
            }
        }

        hr = self.server.additionalHeaders(self.server.timelineHeaders, {
            "x-client-channel": "chat_viewer",
            "x-lal": "zh-Hant_TW",
            "x-recognition-input-type": "obs",
            "Content-Type": "application/json",
        })
        r = self.server.postContent(
            self.LINE_OBS_DOMAIN + '/px/talk/m/%s/recognition.obs' % msgId, data=json.dumps(data), headers=hr)
        if not self.checkRespIsSuccessWithLpv(r):
            raise Exception(f'Training image failure: {r.status_code}')
        return r.json()

    def downloadObjectMsg(self, objId: str, path: str = None, objFrom: str = 'c'):
        obs_path = f'{"g2" if self.getToType(objFrom) == 4 else "talk"}/m'
        return self.downloadObjectForService(objId, path, obs_path)

    def downloadAlbumImage(self, oid: str, path: str):
        obs_path = f'/album/a/download.nhn?ver=1.0&oid={oid}'
        r = self.server.getContent(
            self.LINE_OBS_DOMAIN + obs_path, headers=self.server.timelineHeaders)
        with open(path, 'wb') as f:
            f.write(r.content)
        return r.content

    def downloadProfileImage(self, mid: str, base_path: str, video: bool = False):
        url = self.LINE_OBS_DOMAIN + f'/r/talk/p/{mid}'
        savePath = f"{base_path}/{mid}.jpg"
        if video:
            url += '/vp'
            savePath = f"{base_path}/{mid}.mp4"
        r = self.server.getContent(url, headers=self.server.timelineHeaders)
        with open(savePath, 'wb') as f:
            f.write(r.content)
        return savePath

    def downloadProfileCover(self, mid: str, base_path: str, video: bool = False):
        url, vc_url, objId, vc_objId = self.getProfileCoverObjIdAndUrl(mid)
        savePath = f"{base_path}/{mid}.jpg"
        r = self.server.getContent(url, headers=self.server.timelineHeaders)
        with open(savePath, 'wb') as f:
            f.write(r.content)
        return savePath

    def downloadObjectMyhome(self, objId, path, objFrom='h'):
        return self.downloadObjectForService(objId, path, obsPathPrefix=f'myhome/{objFrom}')

    def downloadImageWithURL(self, url: str, path=None):
        # TEST
        savePath = path
        if savePath is None:
            savePath = f'./.image/{time.time()}.jpg'
        r = self.server.getContant(url)
        if r.status_code != 200:
            raise Exception(f"Not a picture with URL. code: {r.status_code}")
        with open(savePath, 'wb') as f:
            f.write(r.raw)
        return savePath

    def downloadObjectForService(self, objId, savePath, obsPathPrefix='myhome/h', size=None, suffix=None, issueToken4ChannelId: str = None, params: dict = {}):
        obs_path = f'/r/{obsPathPrefix}/{objId}'
        if size is not None:
            # eg. size = "m800x1200" or "L800x1200"or "w800"
            # must be an existing size :p
            obs_path += f'/{size}'
        if suffix is not None:
            # eg. suffix = "mp4"
            obs_path += f'/{suffix}'
        hr = self.Hraders4Obs
        hr = self.server.additionalHeaders(hr, {
            "X-LHM": "GET",
            "X-LPV": "1",
        })
        if issueToken4ChannelId is not None:
            hr = self.server.additionalHeaders(hr, {
                "X-Line-ChannelToken": self.checkAndGetValue(self.approveChannelAndIssueChannelToken(issueToken4ChannelId), 'channelAccessToken', 5),
            })
        isDebugOnly = True
        self.log(f'Starting downloadObjectForService...', isDebugOnly)
        self.log(f'obsPath: {obs_path}', isDebugOnly)
        self.log(f'params: {params}', isDebugOnly)
        self.log(f'hr: {hr}', isDebugOnly)
        try:
            r = self.postPackDataAndGetUnpackRespData(
                f'/oa{obs_path}', b'', 0, 0, headers=hr, conn=self.obsConn)
        except Exception as e:
            print("downloadObjectForService: ", e)
            raise Exception(e)
        if r.status_code != 200:
            raise Exception(
                f'[downloadObjectForService] Download failed! Resp status expected `200`, got `{r.status_code}`.')
        if savePath is not None:
            with open(savePath, 'wb') as f:
                f.write(r.content)
        self.log(f'Ended downloadObjectForService!', isDebugOnly)
        return r.content

    def uploadObjectForService(self, pathOrBytes: any, oType: str = 'image', obsPath: str = 'myhome/h', issueToken4ChannelId: str = None, params: dict = None, talkMeta: str = None, filename: str = None, additionalHeaders: dict = None):
        obs_path = f'/r/{obsPath}'
        hr = self.Hraders4Obs
        data = None
        files = None
        oType = oType.lower()
        isDebugOnly = True
        if isinstance(pathOrBytes, str):
            if pathOrBytes.startswith('http'):
                # URL
                with httpx.Client() as hc:
                    data = hc.get(pathOrBytes).content
            else:
                # FILE
                files = {'file': open(pathOrBytes, 'rb')}
                filename = files['file'].name if filename is None else filename
                filename = filename.split('\\')[-1]
                with open(pathOrBytes, 'rb') as f:
                    data = f.read()
        else:
            # BYTES
            data = pathOrBytes
        filename = str(uuid.uuid1()) if filename is None else filename
        base_params = {
            "type": oType,
            "ver": "2.0",
            "name": filename,
        }
        if params:
            base_params.update(params)
        params = base_params
        if files is not None:
            # FORM DATA
            data = {'params': json.dumps(params)}
            params = None
        else:
            # POST DATA
            files = None
            if len(data) == 0:
                raise ValueError('No data to send: %s' % data)
            hr = self.server.additionalHeaders(hr, {
                'Content-Type': 'image/gif',  # but...
                'Content-Length': str(len(data)),
                'X-Obs-Params': self.genOBSParams(params, 'b64'),
            })
        if issueToken4ChannelId is not None:
            hr = self.server.additionalHeaders(hr, {
                "X-Line-ChannelToken": self.checkAndGetValue(self.approveChannelAndIssueChannelToken(issueToken4ChannelId), 'channelAccessToken', 5),
            })
        if params is not None:
            hr = self.server.additionalHeaders(hr, {
                'X-Obs-Params': self.genOBSParams(params, 'b64').decode(),
            })
        if talkMeta is not None:
            hr = self.server.additionalHeaders(hr, {
                'X-Talk-Meta': talkMeta
            })
        if additionalHeaders is not None:
            hr.update(additionalHeaders)
        self.log(f'Starting uploadObjectForService...', isDebugOnly)
        self.log(f'data: {str(data)[:200]}', isDebugOnly)
        self.log(f'files: {files}', isDebugOnly)
        self.log(f'obsPath: {obs_path}', isDebugOnly)
        self.log(f'params: {params}', isDebugOnly)
        self.log(f'files: {files}', isDebugOnly)
        self.log(f'hr: {hr}', isDebugOnly)
        r = self.postPackDataAndGetUnpackRespData(
            f'/oa{obs_path}', data, 0, 0, headers=hr, expectedRespCode=[200, 201], conn=self.obsConn, files=files)
        #r = self.obsConn.post(self.LINE_GW_HOST_DOMAIN + f'/oa{obs_path}', data=data, headers=hr, files=files)
        # expectedRespCode:
        # - 200: image cache on obs server
        # - 201: image created success
        objId = r.headers['x-obs-oid']
        objHash = r.headers['x-obs-hash']  # for view on cdn
        self.log(f'Ended uploadObjectForService!', isDebugOnly)
        return objId, objHash, r.headers

    def forwardObjectForService(self, data: dict, contentType: str = 'image', copyFromService: str = 'talk', copyFromSid: str = 'm', copyFromObjId: str = None, copyFrom: str = None, copy2Service: str = 'talk', copy2Sid: str = 'm', copy2ObjId: str = None, copy2: str = None, contentId: str = None, original: bool = False, duration: int = None, issueToken4ChannelId: str = None, additionalHeaders: dict = None):
        if contentType.lower() not in ['image', 'video', 'audio', 'file']:
            raise Exception('Type not valid.')
        if copyFrom is None:
            copyFrom = f"/{copyFromService}/{copyFromSid}"
            if copyFromObjId is not None:
                copyFrom += f"/{copyFromObjId}"
        if copy2 is None:
            copy2 = f"{copy2Service}/{copy2Sid}"
            if copy2ObjId is not None:
                copy2 += f"/{copy2ObjId}"
        base_data = {
            'copyFrom': copyFrom,
            'type': contentType,
            'ver': '1.0',
        }
        if contentId is not None:
            base_data['contentId'] = contentId
        if original:
            base_data['cat'] = 'original'
        if duration is not None:
            base_data['duration'] = duration
        base_data.update(data)
        data = base_data
        hr = self.Hraders4Obs
        if issueToken4ChannelId is not None:
            hr = self.server.additionalHeaders(hr, {
                "X-Line-ChannelToken": self.checkAndGetValue(self.approveChannelAndIssueChannelToken(issueToken4ChannelId), 'channelAccessToken', 5),
            })
        if additionalHeaders is not None:
            hr.update(additionalHeaders)
        print(f'data: {data}')
        print(f'hr: {hr}')
        print(f'path: /oa/r/{copy2}/copy.nhn')
        data = list(urllib.parse.urlencode(data).encode())
        r = self.postPackDataAndGetUnpackRespData(f'/oa/r/{copy2}/copy.nhn', data, 0, 0,
                                                  headers=hr,
                                                  conn=self.obsConn)
        objId = r.headers['x-obs-oid']
        objHash = r.headers['x-obs-hash']
        return objId, objHash, r.headers

    def getObjPlayback(self, oid: int, type: str = "VIDEO"):
        url = self.LINE_HOST_DOMAIN + "/oa/talk/m/playback.obs?p=sg-1"
        data = {
            "networkType": "Cell",
            "oid": oid,
            "type": type,
            "ver": "1.0",
            "modelName": "ios",
            "lang": "zh"
        }
        hr = self.server.additionalHeaders(self.server.timelineHeaders, {
            "x-obs-channeltype": "legy",
        })
        r = self.server.postContent(url, data=data, headers=hr)
        if not self.checkRespIsSuccessWithLpv(r):
            raise Exception(f'GetObjPlayback failure: {r.status_code}')
        return r.json()

    def downloadKeepContentByOidAndContentId(self, oid: str, contentId: str, save_path: str = None, sid: str = 'p'):
        return self.downloadObjectForService(
            oid, save_path,
            obsPathPrefix=f'keep/{sid}',
            issueToken4ChannelId='1433572998',
            params={
                'contentId': contentId
            })

    def copyTalkObject4Keep(self, messageId: str, service="talk"):
        url = self.LINE_HOST_DOMAIN + f"/oa/keep/p/copy.nhn"
        hstr = 'DearSakura_%s_%s' % int(time.time() * 1000, messageId)
        file_name = hashlib.md5(hstr.encode()).hexdigest()
        keep_oid = f"linekeep_{file_name}_tffffffff"
        data = {
            'oid': keep_oid,
            'copyFrom': f"/{service}/m/{messageId}"
        }
        hr = self.server.additionalHeaders(self.Hraders4Obs, {
            "x-obs-channeltype": "legy",
        })
        r = self.server.postContent(url, data=data, headers=hr)
        if not self.checkRespIsSuccessWithLpv(r):
            raise Exception(f'saveTalkObject2Keep failure: {r.status_code}')
        return r.json()



import base64
import binascii
import io
import json
import os
import struct
import time
import urllib
from base64 import b64decode, b64encode
from datetime import datetime
from hashlib import md5, sha1

import axolotl_curve25519 as curve
import Crypto.Cipher.PKCS1_OAEP as rsaenc
import httpx
import xxhash
from Crypto.Cipher import AES
from Crypto.PublicKey import RSA
from Crypto.Util.Padding import pad, unpad

from thrift.transport.TTransport import TMemoryBuffer

from .exceptions import LineServiceException
from .serializers.DummyProtocol import DummyProtocol, DummyProtocolData, DummyThrift
from .services.thrift.ap.TBinaryProtocol import \
    TBinaryProtocol as testProtocol2
from .services.thrift.ap.TCompactProtocol import \
    TCompactProtocol as testProtocol

from .services.thrift import *


class Models(object):

    def __init__(self):
        self.lcsStart = "0005"
        self.le = "18"
        self.PUBLIC_KEY = "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA0LRokSkGDo8G5ObFfyKiIdPAU5iOpj+UT+A3AcDxLuePyDt8IVp9HpOsJlf8uVk3Wr9fs+8y7cnF3WiY6Ro526hy3fbWR4HiD0FaIRCOTbgRlsoGNC2rthp2uxYad5up78krSDXNKBab8t1PteCmOq84TpDCRmainaZQN9QxzaSvYWUICVv27Kk97y2j3LS3H64NCqjS88XacAieivELfMr6rT2GutRshKeNSZOUR3YROV4THa77USBQwRI7ZZTe6GUFazpocTN58QY8jFYODzfhdyoiym6rXJNNnUKatiSC/hmzdpX8/h4Y98KaGAZaatLAgPMRCe582q4JwHg7rwIDAQAB\n-----END PUBLIC KEY-----"
        self.key = RSA.importKey(self.PUBLIC_KEY)
        self.encryptKey = b"DearSakura+2021/"
        self.IV = bytes([78, 9, 72, 62, 56, 245, 255, 114,
                        128, 18, 123, 158, 251, 92, 45, 51])
        self.cipher = AES.new(self.encryptKey, AES.MODE_CBC, iv=self.IV)
        self.d_cipher = AES.new(self.encryptKey, AES.MODE_CBC, iv=self.IV)
        self.encEncKey()
        # self.initWithBiz()
        # self.initWithAndroid(4)

    def log(self, text, debugOnly=False):
        if debugOnly and not self.isDebug:
            return
        print(f"[{datetime.now():%Y-%m-%d %H:%M:%S}] {text}")

    def genOBSParams(self, newList, returnAs='json', ext='jpg'):
        oldList = {'name': f'CHRLINE-{int(time.time())}.{ext}', 'ver': '1.0'}
        if returnAs not in ['json', 'b64', 'default']:
            raise Exception('Invalid parameter returnAs')
        oldList.update(newList)
        if 'range' in oldList:
            new_range = 'bytes 0-%s\/%s' % (
                str(oldList['range']-1), str(oldList['range']))
            oldList.update({'range': new_range})
        if returnAs == 'json':
            oldList = json.dumps(oldList)
            return oldList
        elif returnAs == 'b64':
            oldList = json.dumps(oldList)
            return b64encode(oldList.encode('utf-8'))
        elif returnAs == 'default':
            return oldList

    def checkNextToken(self, log4NotDebug: bool = True):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '.tokens')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = md5(self.authToken.encode()).hexdigest()
        if os.path.exists(savePath + f"/{fn}"):
            self.authToken = open(savePath + f"/{fn}", "r").read()
            self.log(f"New Token: {self.authToken}", not log4NotDebug)
            self.checkNextToken(log4NotDebug)
        return self.authToken

    def handleNextToken(self, newToken):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '.tokens')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = md5(self.authToken.encode()).hexdigest()
        open(savePath + f"/{fn}", "w").write(newToken)
        self.authToken = newToken
        self.log(f"New Token: {newToken}")
        self.server.timelineHeaders['X-Line-Access'] = self.authToken
        # need?
        self.server.timelineHeaders['X-Line-ChannelToken'] = self.issueChannelToken()[
            5]

    def getCustomData(self):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '.data')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = md5(self.customDataId.encode()).hexdigest()
        if os.path.exists(savePath + f"/{fn}"):
            self.custom_data = json.loads(
                open(savePath + f"/{fn}", "r").read())
        self.log(f'Loading Custom Data: {fn}')
        return True

    def saveCustomData(self):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '.data')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = md5(self.customDataId.encode()).hexdigest()
        open(savePath + f"/{fn}", "w").write(json.dumps(self.custom_data))
        return True

    def getSqrCert(self):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '.data')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = "cert.pem"
        if os.path.exists(savePath + f"/{fn}"):
            return open(savePath + f"/{fn}", "r").read()
        return None

    def saveSqrCert(self, cert):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '.data')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = "cert.pem"
        open(savePath + f"/{fn}", "w").write(cert)
        return True

    def getEmailCert(self, email):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '.data')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"{email}.crt"
        if os.path.exists(savePath + f"/{fn}"):
            return open(savePath + f"/{fn}", "r").read()
        return None

    def saveEmailCert(self, email, cert):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '.data')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"{email}.crt"
        open(savePath + f"/{fn}", "w").write(cert)
        return True

    def initWithAndroid(self, ver=7):
        if ver == 1:
            self.lcsStart = "0001"
            self.le = "1"
            self.PUBLIC_KEY = '-----BEGIN PUBLIC KEY-----\nMIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQCZAAoZNRwIlLUXaUgrgYi8bAYq\nQeFVtXvCNIEm+F4/jAyTU3YoDwmoLaKQ6itGOonykGtwy2k/3BeWefL/q5eUGjVG\nBEa1vBeUNEb4IFU8n9WK3N/GIIPuD6ZiusB+U1FPg/NaEiVX8ldmEQJgmuG1hykk\n2dU3oy7O1M+Kwl1lJQIDAQAB\n-----END PUBLIC KEY-----'
        elif ver == 4:
            self.lcsStart = "0004"
            self.le = "7"  # LegyEncHelper.cpp::decryptStream -> legy xle value
            self.PUBLIC_KEY = "-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAwpAwTVluR1Z++tVzxtOD\nr7XxSv6oqrwvj/8c8SkfFsS8zM7CvIT8j+x+6Qs1JjNRDtYjAwPKO3tO+qOAdA+8\n7FHpx0THDJIi4VYxSZ2uDh0U8Luxh02whwM8gPbPQNN3sEd5++kJ3cCh5eeAIiUd\nDrwPhHzxO8swpBRdxJB/pzibEqpG2U2764JlPscN9D896qmBN6CBRKpXk/MmUDAI\n4xg+uQk/ykn3SNXJSgQwI1UD9KuiR+X9tbJlKRMN5JpUrSuEwRPQQDMaWpSIdCJM\noFqJLNwt9b1RR/JEB01Eup+3QCub20/CObCmHZY6G26KTDHLoTRZ1xzymdYhdJ43\nCwIDAQAB\n-----END PUBLIC KEY-----"
        elif ver == 6:
            self.lcsStart = "0007"
            self.le = "7"
            self.PUBLIC_KEY = '-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA5ABzJbexh+HH1+RzVTH4\nDFj8b/42vRqUp0NWLIBAgi5+bAgeJYzyVBI7Pk6YkQnd44HPvUvFMF3V3TocRXLP\nZV/NSawgcAh+VrXe3VIlruCOe14qrd/ZMpTRTxtBJ5aRpVhTsnGpZtGggPYnPh4c\n6V/R7Wxymj4SBZ1Ipsa7ZI83z/XIPFXT38qTJN3UAW5YhjQon1eNbwaxALVajuvI\npUE52aIBi05gE/V0HEoOUjfOg1V6RHFbcchTgmcze6Vbye+7kmdsIboDXnNpm/fJ\nuItub+iwLKSWf9OPc/bYpU4YVBxZXzmSCXFMLeCe2i5wJeMg4iG8NpVpwVv2W+Hb\nQQIDAQAB\n-----END PUBLIC KEY-----'
        elif ver == 7:
            self.lcsStart = "0008"
            self.le = "7"
            self.PUBLIC_KEY = '-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAsMC6HAYeMq4R59e2yRw6\nW1OWT2t9aepiAp4fbSCXzRj7A29BOAFAvKlzAub4oxN13Nt8dbcB+ICAufyDnN5N\nd3+vXgDxEXZ/sx2/wuFbC3B3evSNKR4hKcs80suRs8aL6EeWi+bAU2oYIc78Bbqh\nNzx0WCzZSJbMBFw1VlsU/HQ/XdiUufopl5QSa0S246XXmwJmmXRO0v7bNvrxaNV0\ncbviGkOvTlBt1+RerIFHMTw3SwLDnCOolTz3CuE5V2OrPZCmC0nlmPRzwUfxoxxs\n/6qFdpZNoORH/s5mQenSyqPkmH8TBOlHJWPH3eN1k6aZIlK5S54mcUb/oNRRq9wD\n1wIDAQAB\n-----END PUBLIC KEY-----'
        self.key = RSA.importKey(self.PUBLIC_KEY)
        self.IV = bytes([78, 9, 72, 62, 56, 245, 255, 114,
                        128, 18, 123, 158, 251, 92, 45, 51])
        self.encryptKey = b"DearSakura+2021/"
        self.encEncKey()

    def encHeaders(self, headers):
        t = headers.keys()
        data = []
        self.mFhrnmxnNF(len(t), data)
        for i in t:
            self.mFhrnmxnNF(len(i), data)
            self.wYEpEYldst(i, data)
            self.mFhrnmxnNF(len(headers[i]), data)
            self.wYEpEYldst(headers[i], data)
        o = len(data)
        data = [255 & o] + data
        data = [255 & o >> 8] + data
        return data

    def decHeaders(self, data):
        headers = {}
        tbin = self.TBinaryProtocol(self)
        tbin.data = data
        dataLen = tbin.readI16() + 2
        headerLen = tbin.readI16()
        for i in range(headerLen):
            size = tbin.readI16()
            _k = tbin.y(size)
            size = tbin.readI16()
            _v = tbin.y(size)
            headers[_k] = _v
        return headers, data[dataLen:]

    def encEncKey(self):
        # heh
        a = rsaenc.new(self.key)
        self._encryptKey = self.lcsStart + \
            b64encode(a.encrypt(self.encryptKey)).decode()

    def encData(self, data):
        _data = AES.new(self.encryptKey, AES.MODE_CBC,
                        iv=self.IV).encrypt(pad(data, AES.block_size))
        debug = []
        return _data

    def decData(self, data):
        data = pad(data, AES.block_size)
        _data = AES.new(self.encryptKey, AES.MODE_CBC,
                        iv=self.IV).decrypt(data)[:-16]
        _data = unpad(_data, AES.block_size)
        i = 1
        data = self.yVdzCLDwMN(_data, i)
        i = 3
        return _data

    def mFhrnmxnNF(self, t, e):
        i = 65536
        if t < -1 * 32768 or t >= i:
            raise Exception(t + " is incorrect for i16.")
        e.append(255 & t >> 8)
        e.append(255 & t)

    def wYEpEYldst(self, t, e):
        for i in range(len(t)):
            e.append(ord(t[i]))

    def xZVpUuXFru(t):
        if 8 == len(t):
            return t
        e = ""
        i = 0
        n = 8 - len(t)
        while i < n:
            e += "0"
            i += 1
        return e + t

    def pmAWhahfKx(self, t):
        e = []
        i = 0
        n = len(t)
        while i < n:
            _i = 0
            try:
                _i = int(t[i:i + 2], 16)
            except:
                _i = 16
            e.append(_i)
            i += 2
        return e

    def XQqwlHlXKK(self, e, i):
        r = []
        for o in range(16):
            r.append(92 ^ e[o])
        n = xxhash.xxh32(b'', seed=0)
        s = xxhash.xxh32(b'', seed=0)
        n.update(bytes(r))
        for o in range(16):
            r[o] ^= 106
        s.update(bytes(r))
        s.update(i)
        a = s.hexdigest()  # is b8a7c677?
        n.update(bytes(self.pmAWhahfKx(a)))
        c = n.hexdigest()  # is 3f97d2f6?
        d = self.pmAWhahfKx(c)
        return bytes(d)

    def yVdzCLDwMN(self, d, i):
        return (255 & self.xnEmbaRWhy(d, i)) << 8 | 255 & self.xnEmbaRWhy(d, i+1)

    def xnEmbaRWhy(self, d, i):
        t = d[i]
        if t > 127:
            t = 0 - (t - 1 ^ 255)
        return t

    def generateDummyProtocol(self, name, params, type):
        if type == 3:
            data = [128, 1, 0, 1] + self.getStringBytes(name) + [0, 0, 0, 0]
        elif type == 4:
            data = [130, 33, 00] + self.getStringBytes(name, isCompact=True)
        data += self.generateDummyProtocolField(params, type) + [0]
        return data

    def generateDummyProtocol2(self, params: DummyProtocol, type: int = 3, fixSuccessHeaders: bool = False):
        newParams = []
        d = params.data
        if d is not None:
            newParams.append(thrift2dummy(d))
        return bytes(self.generateDummyProtocolField(newParams, type) + [0])

    def generateDummyProtocolField(self, params, type):
        isCompact = False
        data = []
        tcp = self.TCompactProtocol(self)
        for param in params:
            # [10, 2, revision]
            _type = param[0]
            _id = param[1]
            _data = param[2]
            if _data is None:
                continue
            if _type == 13:
                if _data[2] is None:
                    continue
            elif _type in [14, 15]:
                if _data[1] is None:
                    continue
            if type == 3:
                data += [_type, 0, _id]
                isCompact = False
            elif type == 4:
                if _type == 2:
                    data += tcp.getFieldHeader(
                        0x01 if _data else 0x02, _id)
                    continue
                data += tcp.getFieldHeader(tcp.CTYPES[_type], _id)
                isCompact = True
            data += self.generateDummyProtocolData(_data, _type, isCompact)
        return data

    def generateDummyProtocolData(self, _data, type, isCompact=False):
        data = []
        tbp = self.TBinaryProtocol(self)
        tcp = self.TCompactProtocol(self)
        ttype = 4 if isCompact else 3
        if type == 2:
            if isCompact:
                pass  # FIXED
            else:
                data += [1] if _data == True else [0]
        elif type == 3:
            if isCompact:
                data += tcp.writeByte(_data)
            else:
                data += tbp.writeByte(_data)
        elif type == 4:
            data = self.getFloatBytes(_data, isCompact=isCompact)
        elif type == 8:
            data += self.getIntBytes(_data, isCompact=isCompact)
        elif type == 10:
            data += self.getIntBytes(_data, 8, isCompact=isCompact)
        elif type == 11:
            data += self.getStringBytes(_data, isCompact=isCompact)
        elif type == 12:
            if isinstance(_data, DummyProtocolData):
                _data = thrift2dummy(_data)
            data += self.generateDummyProtocolField(_data, ttype) + [0]
        elif type == 13:
            _ktype = _data[0]
            _vtype = _data[1]
            _vdata = _data[2]
            if isCompact:
                data += tcp.writeMapBegin(_ktype, _vtype, len(_vdata))
            else:
                data += [_ktype, _vtype] + \
                    self.getIntBytes(len(_vdata), isCompact=isCompact)
            for vd in _vdata:
                data += self.generateDummyProtocolData(vd, _ktype, isCompact)
                data += self.generateDummyProtocolData(
                    _vdata[vd], _vtype, isCompact)
        elif type == 14 or type == 15:
            # [11, targetUserMids]
            _vtype = _data[0]
            _vdata = _data[1]
            if isCompact:
                data += tcp.writeCollectionBegin(_vtype, len(_vdata))
            else:
                data += [_vtype] + \
                    self.getIntBytes(len(_vdata), isCompact=isCompact)
            for vd in _vdata:
                data += self.generateDummyProtocolData(vd, _vtype, isCompact)
        else:
            raise Exception(
                f"[generateDummyProtocolData] not support type: {type}")
        return data

    def postPackDataAndGetUnpackRespData(self, path: str, bdata: bytes, ttype: int = 3, encType: int = None, headers: dict = None, access_token: str = None, baseException: dict = None, readWith: str = None, conn: any = None, files: dict = None, expectedRespCode: list = [200], timeout: int = None):
        if headers is None:
            headers = self.server.Headers.copy()
        if access_token is None:
            access_token = self.authToken
        ptype = "TBINARY" if ttype == 3 else "TCOMPACT"
        if ttype in [1, 2, 3, 4, 5]:
            headers["content-type"] = "application/x-thrift; protocol=" + ptype
            headers["accept"] = "application/x-thrift"
        headers["x-lal"] = self.LINE_LANGUAGE
        if encType is None:
            encType = self.encType
        self.log(
            f"--> POST {path} {f'({self.LINE_ENCRYPTION_ENDPOINT})' if encType == 1 else ''}", True)
        if encType == 0:
            if conn is None:
                conn = self.req_h2
            data = bytes(bdata) if isinstance(bdata, list) else bdata
            if "x-le" in headers:
                del headers['x-le']
                del headers['x-lcs']
            if access_token is not None:
                headers['X-Line-Access'] = access_token
            self.log(f"--> Headers: {headers}", True)
            res = doLoopReq(conn.post, {
                "url": self.LINE_GW_HOST_DOMAIN + path,
                "data": data,
                "headers": headers, 
                "files": files,
                "timeout": timeout
            })
            data = res.content
        elif encType == 1:
            if conn is None:
                conn = self.req
            if access_token is not None:
                _headers = {
                    'x-lt': access_token,
                    'x-lpqs': path
                }
            else:
                _headers = {
                    'x-lpqs': path
                }
            a = self.encHeaders(_headers)
            b = bdata
            c = a + b
            _data = bytes(c)
            fix_bytes = False
            if ((int(self.le) & 4) == 4):
                _data = bytes([int(self.le)]) + _data
                fix_bytes = True
            if ((int(self.le) & 2) != 2):
                data = self.encData(_data)
            else:
                data = self.encData(_data)
                data += self.XQqwlHlXKK(self.encryptKey, data)
            headers['accept-encoding'] = 'gzip, deflate'
            self.log(f"--> Headers: {headers} ({_headers})", True)
            res = doLoopReq(conn.post, {
                "url": self.LINE_GF_HOST_DOMAIN + self.LINE_ENCRYPTION_ENDPOINT,
                "data": data, 
                "files": files, 
                "headers": headers,
                "timeout": timeout
            })
            if res.content:
                data = self.decData(res.content)
            else:
                data = res.content
            if fix_bytes:
                data = data[1:]
        else:
            raise Exception(f"Unknown encType: {encType}")
        self.log(f"<--  {res.status_code}", True)
        self.log(f"{data}", True)
        if res.status_code in expectedRespCode:
            if res.headers.get('x-lc') is not None and int(res.headers['x-lc']) not in expectedRespCode:
                raise Exception(
                    f'Invalid response code: {res.headers["x-lc"]}')
            if encType == 1:
                respHeaders, data = self.decHeaders(data)
            else:
                respHeaders = {}
            respHeaders.update(res.headers)
            self.log(f"RespHraders: {respHeaders}", True)
            if 'x-line-next-access' in respHeaders:
                print(respHeaders)
                self.handleNextToken(respHeaders['x-line-next-access'])
            conn_res = res
            res = None
            if ttype == -7:
                # COMPACT
                # TODO: ADD DECODER
                compact = self.TCompactProtocol(self)
                _type = compact.readByte(data)
                data = data[1:]
                if _type == 1:
                    _seq, _offset = compact.readI32(data, True)
                    data = data[_offset:]
                    _msgId, _offset = compact.readI32(data, True)
                    data = data[_offset:]
                    _ts, _offset = compact.readI32(data, True)
                    data = data[_offset:]
                    compact.res = {
                        '_seq': _seq,
                        'messageId': _msgId,
                        'time': _ts
                    }
                    res = compact
                elif _type == 2:
                    raise LineServiceException({'code': compact.readI32(data)})
            elif ttype == -3:
                # JSON RAW
                return json.loads(data)
            elif ttype == 0:
                # RESP
                return conn_res
            elif ttype == 3:
                res = self.TBinaryProtocol(
                    self, data, baseException=baseException)
            elif ttype == 4:
                res = self.TCompactProtocol(
                    self, data, baseException=baseException)
            elif ttype == 5:
                tmore = self.TMoreCompactProtocol(
                    self, data, baseException=baseException)
                res = tmore
            else:
                raise ValueError(f"Unknown ThriftType: {ttype}")
            if self.use_thrift and getattr(res, 'dummyProtocol', None) is not None:
                res = self.serializeDummyProtocolToThrift(
                    res.dummyProtocol, baseException, readWith)
            else:
                res = res.res
            if type(res) == dict and 'error' in res:
                if res['error']['message'] is not None and (res['error']['message'] in ["EXPIRED", "REVOKE", "LOG_OUT", "AUTHENTICATION_DIVESTED_BY_OTHER_DEVICE", "DEVICE_LOSE", "IDENTIFY_MODIFIED", "V3_TOKEN_CLIENT_LOGGED_OUT", "DELETED_ACCOUNT"] or res['error']['message'].startswith('suspended')):
                    self.is_login = False
                    self.log(f"LOGIN OUT: {res['error']['message']}")
                elif res['error']['code'] == 119:
                    refreshToken = self.getCacheData(
                        '.refreshToken', self.authToken)
                    print(f'try to refresh access token... {refreshToken}')
                    if refreshToken is not None:
                        RATR = self.refreshAccessToken(refreshToken)
                        token = self.checkAndGetValue(RATR, 'accessToken', 1)
                        # refreshToken = self.checkAndGetValue(RATR, 'refreshToken', 5)
                        self.handleNextToken(token)
                        self.saveCacheData(
                            '.refreshToken', token, refreshToken)
                        return self.postPackDataAndGetUnpackRespData(path, bdata, ttype, encType, headers)
                    self.log(f"LOGIN OUT: {res['error']['message']}")
                raise LineServiceException(res['error'])
            return res
        elif res.status_code in [400, 401, 403]:
            self.is_login = False
        elif res.status_code == 410:
            return None
        raise Exception(f'Invalid response status code: {res.status_code}')

    def getCurrReqId(self):
        self._msgSeq = 0
        if "_reqseq" in self.custom_data:
            self._msgSeq = self.custom_data["_reqseq"]
        self._msgSeq += 1
        self.custom_data["_reqseq"] = self._msgSeq
        self.saveCustomData()
        return self._msgSeq

    def getIntBytes(self, i, l=4, isCompact=False):
        i = int(i)
        if isCompact:
            _compact = self.TCompactProtocol(self)
            a = _compact.makeZigZag(i, 32 if l**2 == 16 else 64)
            b = _compact.writeVarint(a)
            return b
        if l**2 == 16:
            res = struct.pack("!i", i)
        else:
            res = struct.pack("!q", i)
        return list(res)

    def getStringBytes(self, text, isCompact=False):
        if text is None:
            text = ""
        if type(text) == bytes:
            pass
        else:
            text = str(text).encode()
        if isCompact:
            _compact = self.TCompactProtocol(self)
            sqrd = _compact.writeVarint(len(text))
        else:
            sqrd = self.getIntBytes(len(text))
        for value in text:
            sqrd.append(value)
        return sqrd

    def getFloatBytes(self, val, isCompact=False):
        res = []
        _t = "!d"
        if isCompact:
            _t = "<d"
        for value in struct.pack(_t, val):
            res.append(value)
        return res

    def getMagicStringBytes(self, val, rev=False):
        res = []
        i = 0
        if rev:
            res = binascii.b2a_hex(val)
        else:
            if len(val) == 32:
                for ii in range(16):
                    iii = ii * 2
                    i = iii + 1
                    mgc = (int(val[iii], 16) << 4) + int(val[i], 16)
                    res.append(mgc)
            else:
                raise ValueError(
                    f"getMagicStringBytes() expected 32, but got {len(val)}")
        return res

    def createSqrSecret(self, base64Only=False):
        private_key = curve.generatePrivateKey(os.urandom(32))
        public_key = curve.generatePublicKey(private_key)
        secret = urllib.parse.quote(b64encode(public_key).decode())
        version = 1
        if base64Only:
            return [private_key, b64encode(public_key).decode()]
        return [private_key, f"?secret={secret}&e2eeVersion={version}"]

    def getE2EESelfKeyData(self, mid):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '.e2eeKeys')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"{mid}.json"
        if os.path.exists(savePath + f"/{fn}"):
            return json.loads(open(savePath + f"/{fn}", "r").read())
        print(savePath + f"/{fn}")
        keys = self.getE2EEPublicKeys()
        for key in keys:
            keyId = self.checkAndGetValue(key, 'keyId', 2)
            _keyData = self.getE2EESelfKeyDataByKeyId(keyId)
            if _keyData is not None:
                return _keyData
        raise Exception(
            'E2EE Key has not been saved, try register or use SQR Login')

    def getE2EESelfKeyData(self, mid):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '.e2eeKeys')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"{mid}.json"
        if os.path.exists(savePath + f"/{fn}"):
            return json.loads(open(savePath + f"/{fn}", "r").read())
        keys = self.getE2EEPublicKeys()
        for key in keys:
            _keyData = self.getE2EESelfKeyDataByKeyId(key[2])
            if _keyData is not None:
                return _keyData
        return None

    def getE2EESelfKeyDataByKeyId(self, keyId):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '.e2eeKeys')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"key_{keyId}.json"
        if os.path.exists(savePath + f"/{fn}"):
            return json.loads(open(savePath + f"/{fn}", "r").read())
        return None

    def saveE2EESelfKeyData(self, mid, pubK, privK, kI, e2eeVersion):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), '.e2eeKeys')
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"{mid}.json"
        fn2 = f"key_{kI}.json"
        data = json.dumps({
            "keyId": kI,
            "privKey": b64encode(privK).decode(),
            "pubKey": b64encode(pubK).decode(),
            "e2eeVersion": e2eeVersion,
        })
        if mid is not None:
            open(savePath + f"/{fn}", "w").write(data)
        open(savePath + f"/{fn2}", "w").write(data)
        return True

    def registerE2EESelfKey(self, privK: bytes = None):
        if privK is None:
            privK = curve.generatePrivateKey(os.urandom(32))
        if len(privK) != 32:
            raise ValueError("Invalid private key.")
        pubK = curve.generatePublicKey(privK)
        EPK = self.registerE2EEPublicKey(1, None, pubK, 0)
        keyId = self.checkAndGetValue(EPK, 'keyId', 2)
        return self.saveE2EESelfKeyData(self.mid, pubK, privK, keyId, 1)

    def getCacheData(self, cT, cN, needHash=True, pathOnly=False):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), cT)
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"{cN}"
        if needHash:
            fn = md5(cN.encode()).hexdigest()
        if pathOnly:
            return savePath + f"/{fn}"
        if os.path.exists(savePath + f"/{fn}"):
            return open(savePath + f"/{fn}", "r").read()
        return None

    def saveCacheData(self, cT, cN, cD, needHash=True):
        savePath = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), cT)
        if not os.path.exists(savePath):
            os.makedirs(savePath)
        fn = f"{cN}"
        if needHash:
            fn = md5(cN.encode()).hexdigest()
        data = cD
        open(savePath + f"/{fn}", "w").write(data)
        return True

    def decodeE2EEKeyV1(self, data, secret, mid=None):
        if 'encryptedKeyChain' in data:
            print("Try to decode E2EE Key")
            encryptedKeyChain = base64.b64decode(data['encryptedKeyChain'])
            # hashKeyChain = data['hashKeyChain']
            keyId = data['keyId']
            publicKey = base64.b64decode(data['publicKey'])
            e2eeVersion = data['e2eeVersion']
            e2eeKey = self.decryptKeyChain(
                publicKey, secret, encryptedKeyChain)
            print(f"E2EE Priv Key: {e2eeKey[0]}")
            print(f"E2EE Pub Key: {e2eeKey[1]}")
            print(f"keyId: {keyId}")
            print(f"e2eeVersion: {e2eeVersion}")
            self.saveE2EESelfKeyData(
                mid, e2eeKey[1], e2eeKey[0], keyId, e2eeVersion)
            return {
                "keyId": keyId,
                "privKey": e2eeKey[0],
                "pubKey": e2eeKey[1],
                "e2eeVersion": e2eeVersion,
            }

    def tryReadThriftContainerStruct(self, data, id=0, get_data_len=False):
        _data = {}
        _dec = self.TCompactProtocol(self)
        ftype = data[0] & 15
        fid = (data[0] >> 4) + id
        offset = 1
        nextPos = 0
        if ftype == 0:
            _data = None
        elif ftype == 4:
            _data[fid] = _dec.readDouble(data[offset:])
            nextPos += 8
        elif ftype == 5:
            (_data[fid], nextPos) = _dec.readI32(data[offset:], True)
            nextPos += 1
        elif ftype == 6:
            (_data[fid], nextPos) = _dec.readI64(data[offset:], True)
            nextPos += 1
        elif ftype == 8:
            (_data[fid], nextPos) = _dec.readBinary(data[offset:])
        elif ftype == 9 or ftype == 10:
            (vtype, vsize, vlen) = _dec.readCollectionBegin(data[offset:])
            offset += vlen
            _data[fid] = []
            _nextPos = 0
            for i in range(vsize):
                if vtype == 12:
                    _aaa, _bbb = self.tryReadThriftContainerStruct(
                        data[offset:], get_data_len=True)
                    _data[fid].append(_aaa)
                    offset += _bbb + 1
        else:
            print(
                f"[tryReadThriftContainerStruct]不支援Type: {ftype} => ID: {fid}")
        if nextPos > 0:
            data = data[nextPos:]
            c = self.tryReadThriftContainerStruct(
                data, id=fid, get_data_len=True)
            if c[0] is not None:
                _data.update(c[0])
            nextPos += c[1]
        if get_data_len:
            return [_data, nextPos]
        return _data

    def serializeDummyProtocolToThrift(self, data: DummyProtocol, baseException: dict = None, readWith: str = None):
        if baseException is None:
            baseException = {}
        if readWith is not None:
            new1 = self.generateDummyProtocol2(data, 4, fixSuccessHeaders=True)
            try:
                a = eval(f'{readWith}_result')
                a = a()
            except AttributeError:
                a = None
            except NameError:
                a = None
            if a is not None:
                e = TMemoryBuffer()
                f = testProtocol(e)
                e._buffer = io.BytesIO(new1)
                a.read(f)
                if getattr(a, 'success', None) is not None:
                    return a.success
                if getattr(a, 'e', None) is not None:
                    code = getattr(a.e, 'code', None)
                    reason = getattr(a.e, 'reason', None)
                    parameterMap = getattr(a.e, 'parameterMap', None)
                    raise LineServiceException(
                        {}, code, reason, parameterMap, a.e)
                return None

        def _gen(): return DummyThrift()

        def _get(a): return a.data if isinstance(a, DummyProtocolData) else a

        def _genFunc(a: DummyProtocolData, b, f):
            def __gen(a: DummyProtocolData, b):
                c = _gen()
                for d in a.data:
                    b(d, c)
                return c

            def __cek(a: DummyProtocolData, f):
                if a.type == 12:
                    c = __gen(a, f)
                elif a.type == 13:
                    c = {}
                    d = a.data
                    for e in d:
                        g = d[e]
                        h = e
                        if isinstance(h, DummyProtocolData):
                            h = __cek(h, f)
                        if isinstance(g, DummyProtocolData):
                            g = __cek(g, f)
                        c[h] = g
                elif a.type in (14, 15):
                    c = []
                    for d in a.data:
                        e = d
                        if isinstance(d, DummyProtocolData):
                            e = __cek(d, f)
                        c.append(e)
                else:
                    c = a.data
                return c
            c = __cek(a, f)
            setattr(b, f"val_{a.id}", c)
        a = _gen()

        def b(c, refs): 
            return _genFunc(c, refs, b) if type(c.data) in [
            list, dict] else setattr(refs, f"val_{c.id}", c.data)
        if data.data is not None:
            b(data.data, a)
        if self.checkAndGetValue(a, 'val_0') is not None:
            return a.val_0
        _ecode = baseException.get('code', 1)
        _emsg = baseException.get('message', 2)
        _emeta = baseException.get('metadata', 3)
        if self.checkAndGetValue(a, 'val_1') is not None:
            raise LineServiceException({}, self.checkAndGetValue(a.val_1, f'val_{_ecode}'), self.checkAndGetValue(
                a.val_1, f'val_{_emsg}'), self.checkAndGetValue(a.val_1, f'val_{_emeta}'), a.val_1)
        print(a)
        return None


def thrift2dummy(a):
    if type(a) == dict:
        b = {}
        for k, v in a.items():
            b[thrift2dummy(k)] = thrift2dummy(v)
        return b
    elif type(a) == list:
        return [thrift2dummy(a2) for a2 in a]
    elif isinstance(a, DummyProtocolData):
        b = None
        c = []
        if a.type in [2, 3, 4, 6, 8, 10, 11]:
            b = a.data
        elif a.type == 12:
            b = [thrift2dummy(a2) for a2 in a.data]
        elif a.type == 13:
            b = [a.subType[0], a.subType[1], thrift2dummy(a.data)]
        elif a.type in [14, 15]:
            b = [a.subType[0], thrift2dummy(a.data)]
        else:
            raise ValueError(f"Not supported type: {a.type}")
        if a.id is not None:
            return [a.type, a.id, b]
        return b
        return [a.type, a.id, a.data] if a.type in [2, 3, 4, 6, 8, 10, 11] else (a.subType + [[thrift2dummy(a2) for a2 in a.data]] if a.subType else [thrift2dummy(a2) for a2 in a.data]) if a.id is None else [a.type, a.id, [thrift2dummy(a2) for a2 in a.data]] if a.type in [12] else [a.type, a.id, [a.subType[0], a.subType[1], thrift2dummy(a.data)]] if a.type == 13 else [a.type, a.id, [a.subType[0], [thrift2dummy(a2) if isinstance(a2, DummyProtocolData) and a2.type == 12 else a2.data for a2 in a.data]]] if a.type in [14, 15] else (_ for _ in ()).throw(ValueError(f"不支持{a.type}"))
    else:
        # return a
        raise ValueError(f"不支持 `{type(a)}`: {a}")

def doLoopReq(req, data, currCount: int = 0, maxRetryCount: int = 5, retryTimeDelay: int = 8):
    currCount += 1
    doRetry = False
    e = None
    try:
        res = req(**data)
    except httpx.ConnectTimeout as ex:
        doRetry = True
        e = ex
    except httpx.ReadTimeout as ex:
        currCount -= 1
        doRetry = True
        e = ex
    except httpx.ReadError as ex:
        doRetry = True
        e = ex
    except httpx.ConnectError as ex:
        doRetry = True
        e = ex
    if doRetry:
        if currCount > maxRetryCount:
            raise e
        time.sleep(retryTimeDelay)
        return doLoopReq(req, data, currCount, maxRetryCount, retryTimeDelay)
    return res




class DummyProtocolData():

    def __init__(self, id, type, data, subType: list = None):
        self.id = id
        self.type = type
        self.data = data
        self.subType = []
        if subType is not None:
            for _subType in subType:
                self.addSubType(_subType)

    def addSubType(self, type):
        self.subType.append(type)

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

class DummyProtocol():

    def __init__(self, type: int=5, data: DummyProtocolData= None):
        self.type = 5
        self.data = data
    
    def dump(self):
        data = b''
        data += bytes([130, 33, 0, 0])
        return data

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

class DummyThrift():

    def __init__(self, name: str = None, **kwargs):
        if name is not None:
            self.__name__ = name
        if kwargs:
            for key in kwargs:
                setattr(self, key, kwargs[key])

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))


# -*- coding: utf-8 -*-

class HomeSafetyCheckService(object):
    HomeSafetyCheckService_REQ_TYPE = 4
    HomeSafetyCheckService_RES_TYPE = 4
    HomeSafetyCheckService_API_PATH = "/EXT/home/safety-check/safety-check"

    def __init__(self):
        pass

    def deleteSafetyStatus(self, disasterId: str):
        params = [
            [12, 1, [
                [11, 1, disasterId]
            ]]
        ]
        sqrd = self.generateDummyProtocol(
            'deleteSafetyStatus', params, self.HomeSafetyCheckService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.HomeSafetyCheckService_API_PATH, sqrd, self.HomeSafetyCheckService_RES_TYPE)

    def getDisasterCases(self):
        params = [
            [12, 1, [
                []
            ]]
        ]
        sqrd = self.generateDummyProtocol(
            'getDisasterCases', params, self.HomeSafetyCheckService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.HomeSafetyCheckService_API_PATH, sqrd, self.HomeSafetyCheckService_RES_TYPE)

    def updateSafetyStatus(self, disasterId: str, message: str, safetyStatus: int = 1):
        """
        - safetyStatus:
            SAFE(1),
            NOT_SAFE(2);
        """
        params = [
            [12, 1, [
                [11, 1, disasterId],
                [8, 2, safetyStatus],
                [11, 3, message],
            ]]
        ]
        sqrd = self.generateDummyProtocol(
            'updateSafetyStatus', params, self.HomeSafetyCheckService_REQ_TYPE)
        return self.postPackDataAndGetUnpackRespData(self.HomeSafetyCheckService_API_PATH, sqrd, self.HomeSafetyCheckService_RES_TYPE)
